/**
 * \file pappsomspp/msrun/private/timsmsrunreader.h
 * \date 05/09/2019
 * \author Olivier Langella
 * \brief MSrun file reader for native Bruker TimsTOF raw data
 */

/*******************************************************************************
 * Copyright (c) 2019 Olivier Langella <Olivier.Langella@u-psud.fr>.
 *
 * This file is part of the PAPPSOms++ library.
 *
 *     PAPPSOms++ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     PAPPSOms++ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with PAPPSOms++.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

#include "timsmsrunreader.h"
#include "../../exception/exceptionnotimplemented.h"
#include "../../exception/exceptioninterrupted.h"
#include <QDebug>

using namespace pappso;

TimsMsRunReader::TimsMsRunReader(MsRunIdCstSPtr &msrun_id_csp)
  : MsRunReader(msrun_id_csp)
{
  initialize();
}

TimsMsRunReader::~TimsMsRunReader()
{
  msp_timsData = nullptr;
}

void
TimsMsRunReader::initialize()
{
  msp_timsData = std::make_shared<TimsData>(mcsp_msRunId.get()->getFileName());

  if(msp_timsData == nullptr)
    {
      throw PappsoException(QObject::tr("ERROR in TimsMsRunReader::initialize "
                                        "msp_timsData is null for MsRunId %1")
                              .arg(mcsp_msRunId.get()->toString()));
    }
}


bool
TimsMsRunReader::accept(const QString &file_name) const
{
  qDebug() << file_name;
  return true;
}


pappso::MassSpectrumSPtr
TimsMsRunReader::massSpectrumSPtr([[maybe_unused]] std::size_t spectrum_index)
{
  throw ExceptionNotImplemented(
    QObject::tr("Not yet implemented in TimsMsRunReader %1.\n").arg(__LINE__));
  return pappso::MassSpectrumSPtr();
}


pappso::MassSpectrumCstSPtr
TimsMsRunReader::massSpectrumCstSPtr(std::size_t spectrum_index)
{
  return msp_timsData->getMassSpectrumCstSPtrByRawIndex(spectrum_index);
}


QualifiedMassSpectrum
TimsMsRunReader::qualifiedMassSpectrum(std::size_t spectrum_index,
                                       bool want_binary_data) const
{

  QualifiedMassSpectrum mass_spectrum;

  msp_timsData->getQualifiedMassSpectrumByRawIndex(
    getMsRunId(), mass_spectrum, spectrum_index, want_binary_data);
  return mass_spectrum;
}


void
TimsMsRunReader::readSpectrumCollection(
  SpectrumCollectionHandlerInterface &handler)
{
  readSpectrumCollectionByMsLevel(handler, 0);
}

void
TimsMsRunReader::readSpectrumCollection2(
  const MsRunReadConfig &config, SpectrumCollectionHandlerInterface &handler)
{

  qDebug().noquote() << "Reading the spectrum collection with this "
                        "specific configuration:"
                     << config.toString();

  // We want to restrict the data reading process to the configuration provided
  // as parameter.
  // The frame record subset will include all frames fulfiling the config
  // requirements (retention time and ms levels)

  std::vector<std::size_t> subset_of_tims_frame_ids;

  // Check for ion mobility range :
  bool asked_ion_mobility_range   = false;
  quint32 mobility_scan_num_begin = 0;
  quint32 mobility_scan_num_end   = 0;

  if(!config
        .getParameterValue(
          MsRunReadConfigParameter::TimsFramesMsRunReader_mobility_index_begin)
        .isNull())
    {
      asked_ion_mobility_range = true;

      mobility_scan_num_begin =
        config
          .getParameterValue(MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mobility_index_begin)
          .toUInt();
      mobility_scan_num_end =
        config
          .getParameterValue(
            MsRunReadConfigParameter::TimsFramesMsRunReader_mobility_index_end)
          .toUInt();
    }


  // Now that we have gathered all the configuration bits from the user,
  // iterate in the vector of frames records and for each frame ask that
  // all its scans be combined into a single mass spectrum, by taking into
  // account the configuration bits provided by the user.

  const std::vector<FrameIdDescr> &frame_id_descr_list =
    msp_timsData->getFrameIdDescrList();

  // The very first sorting of frames is perfomed on the basis of the
  // retention time range that the user might have defined. Otherwise, all the
  // frames are deemed suitable for next processing steps.

  // Then, the MS level is checked against the config provided by the user.
  std::size_t scan_count = 0;
  for(auto const &frame_record : msp_timsData->getTimsFrameRecordList())
    {
      if(frame_record.frame_id == 0)
        continue;

      if(config.acceptRetentionTimeInSeconds(frame_record.frame_time))
        {
          std::size_t ms_level = 2;
          if(frame_record.msms_type == 0)
            ms_level = 1;
          if(config.acceptMsLevel(ms_level))
            {
              subset_of_tims_frame_ids.push_back(frame_record.frame_id);

              if(mobility_scan_num_end == 0)
                {
                  scan_count +=
                    frame_id_descr_list[frame_record.frame_id].m_size;
                }
              else
                {
                  scan_count +=
                    mobility_scan_num_end + 1 - mobility_scan_num_begin;
                }
            }
        }
    }


  // Inform the handler of the spectrum list so that it can handle feedback to
  // the user.
  handler.spectrumListHasSize(scan_count);

  // Check for m/z resolution downgrading
  std::size_t mz_index_merge_window = 0;
  if(!config
        .getParameterValue(
          MsRunReadConfigParameter::TimsFramesMsRunReader_mz_index_merge_window)
        .isNull())
    {
      mz_index_merge_window =
        config
          .getParameterValue(MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mz_index_merge_window)
          .toUInt();

      qDebug() << "mz_index_merge_window=" << mz_index_merge_window;
    }

  // Check for m/z range selection
  bool asked_mz_range   = false;
  double mz_range_begin = -1;
  double mz_range_end   = -1;

  if(!config
        .getParameterValue(
          MsRunReadConfigParameter::TimsFramesMsRunReader_mz_begin)
        .isNull())
    {
      asked_mz_range = true;

      mz_range_begin =
        config
          .getParameterValue(
            MsRunReadConfigParameter::TimsFramesMsRunReader_mz_begin)
          .toDouble();

      mz_range_end = config
                       .getParameterValue(
                         MsRunReadConfigParameter::TimsFramesMsRunReader_mz_end)
                       .toDouble();

      qDebug() << "The m/z range asked is: " << mz_range_begin << "--"
               << mz_range_end;
    }


  // The scan index is the index of the scan in the *whole* mass data file, it
  // is a sequential number of scans over all the frames.
  std::size_t scan_index = 0; // iterate in each spectrum

  for(std::size_t tims_frame_id : subset_of_tims_frame_ids)
    {
      qDebug() << "tims_frame_id=" << tims_frame_id;
      const FrameIdDescr &current_frame_record =
        frame_id_descr_list[tims_frame_id];
      qDebug() << "tims_frame_id=" << tims_frame_id;

      TimsFrameCstSPtr tims_frame_csp =
        msp_timsData->getTimsFrameCstSPtrCached(tims_frame_id);
      qDebug() << "tims_frame_id=" << tims_frame_id;

      // If the user wants to select specific 1/Ko values, we need to
      // compute the ion mobility scan value starting from that 1/Ko value in
      // each frame.

      if(!config
            .getParameterValue(
              MsRunReadConfigParameter::
                TimsFramesMsRunReader_mobility_one_over_k0_begin)
            .isNull())
        {
          asked_ion_mobility_range = true;

          mobility_scan_num_begin =
            tims_frame_csp.get()->getScanNumFromOneOverK0(
              config
                .getParameterValue(
                  MsRunReadConfigParameter::
                    TimsFramesMsRunReader_mobility_one_over_k0_begin)
                .toDouble());

          mobility_scan_num_end = tims_frame_csp.get()->getScanNumFromOneOverK0(
            config
              .getParameterValue(
                MsRunReadConfigParameter::
                  TimsFramesMsRunReader_mobility_one_over_k0_end)
              .toDouble());
        }
      else if(!config
                 .getParameterValue(
                   MsRunReadConfigParameter::
                     TimsFramesMsRunReader_mobility_index_begin)
                 .isNull() &&
              !config
                 .getParameterValue(MsRunReadConfigParameter::
                                      TimsFramesMsRunReader_mobility_index_end)
                 .isNull())
        {
          asked_ion_mobility_range = true;

          mobility_scan_num_begin =
            config
              .getParameterValue(MsRunReadConfigParameter::
                                   TimsFramesMsRunReader_mobility_index_begin)
              .toDouble();

          mobility_scan_num_end =
            config
              .getParameterValue(MsRunReadConfigParameter::
                                   TimsFramesMsRunReader_mobility_index_end)
              .toDouble();
        }

      qDebug() << "tims_frame_id=" << tims_frame_id;

      // Get to know the size of the frame, that is,
      // the number of mobility scans in it.

      quint32 mobility_scans_count = tims_frame_csp->getTotalNumberOfScans();

      // Provide these two variables for the function below to fill in the
      // values.
      quint32 min_mz_index_out = 0;
      quint32 max_mz_index_out = 0;

      qDebug() << "tims_frame_id=" << tims_frame_id;
      if(asked_ion_mobility_range)
        {
          if(mobility_scan_num_end > (mobility_scans_count - 1))
            {
              mobility_scan_num_end = mobility_scans_count - 1;
            }
        }
      else
        {
          mobility_scan_num_begin = 0;
          mobility_scan_num_end   = mobility_scans_count - 1;
        }

      qDebug() << "tims_frame_id=" << tims_frame_id;
      // qDebug() << "Got min_mz_index_out:" << min_mz_index_out;
      // qDebug() << "Got max_mz_index_out:" << max_mz_index_out;

      // Now, with or without the peak list, we have to craft a qualified mass
      // spectrum that will hold all the data about the data in it.
      QualifiedMassSpectrum mass_spectrum;

      MassSpectrumId spectrum_id;

      spectrum_id.setSpectrumIndex(tims_frame_id);
      spectrum_id.setMsRunId(getMsRunId());

      mass_spectrum.setMassSpectrumId(spectrum_id);

      // We want to document the retention time!
      mass_spectrum.setRtInSeconds(tims_frame_csp.get()->getTime());

      // We do want to document the ms level of the spectrum and possibly
      // the precursor's m/z and charge.
      unsigned int frame_ms_level = tims_frame_csp.get()->getMsLevel();
      mass_spectrum.setMsLevel(frame_ms_level);


      // qDebug() << "Setting mass spectrum parameter value mz index begin to:"
      //          << min_mz_index_out;
      mass_spectrum.setParameterValue(
        QualifiedMassSpectrumParameter::TimsFrameMzIndexBegin,
        min_mz_index_out);

      // qDebug() << "Setting mass spectrum parameter value mz index end to:"
      //          << max_mz_index_out;

      scan_index = current_frame_record.m_cumulSize -
                   current_frame_record.m_size + mobility_scan_num_begin;
      for(quint32 i = mobility_scan_num_begin; i <= mobility_scan_num_end; i++)
        {
          mass_spectrum.getMassSpectrumId().setSpectrumIndex(scan_index);

          mass_spectrum.getMassSpectrumId().setNativeId(
            QString("frame=%1 scan=%2 index=%3")
              .arg(tims_frame_id)
              .arg(i)
              .arg(scan_index));

          qDebug() << "i=" << i;

          if(config.needPeakList())
            {
              quint32 mz_minimum_index_out =
                std::numeric_limits<quint32>::max();
              quint32 mz_maximum_index_out = 0;
              auto raw =
                tims_frame_csp.get()->getMobilityScan(i,
                                                      mz_index_merge_window,
                                                      mz_range_begin,
                                                      mz_range_end,
                                                      mz_minimum_index_out,
                                                      mz_maximum_index_out);
              qDebug() << raw.size();
              mass_spectrum.setEmptyMassSpectrum(false);


              mass_spectrum.setDtInMilliSeconds(
                tims_frame_csp.get()->getDriftTime(i));
              // 1/K0
              mass_spectrum.setParameterValue(
                QualifiedMassSpectrumParameter::OneOverK0,
                tims_frame_csp.get()->getOneOverK0Transformation(i));

              /*
                    mass_spectrum.setParameterValue(
                      QualifiedMassSpectrumParameter::TimsFrameMzIndexEnd,
                 max_mz_index_out);
                 */
              qDebug();
              mass_spectrum.setMassSpectrumSPtr(
                std::make_shared<MassSpectrum>(raw));

              qDebug();
            }
          else
            {
              mass_spectrum.setEmptyMassSpectrum(true);
            }

          qDebug() << "mz_index_merge_window=" << mz_index_merge_window;
          handler.setQualifiedMassSpectrum(mass_spectrum);
          scan_index++;
        }
    }
}

void
TimsMsRunReader::readSpectrumCollectionByMsLevel(
  SpectrumCollectionHandlerInterface &handler, unsigned int ms_level)
{

  qDebug();

  try
    {

      msp_timsData.get()->rawReaderSpectrumCollectionByMsLevel(
        getMsRunId(), handler, ms_level);
    }

  catch(ExceptionInterrupted &)
    {
      qDebug() << "Reading of MS data interrupted by the user.";
    }

  // Now let the loading handler know that the loading of the data has ended.
  // The handler might need this "signal" to perform additional tasks or to
  // cleanup cruft.

  // qDebug() << "Loading ended";
  handler.loadingEnded();
}


std::size_t
TimsMsRunReader::spectrumListSize() const
{
  return msp_timsData->getTotalNumberOfScans();
}


bool
TimsMsRunReader::hasScanNumbers() const
{
  return false;
}


bool
TimsMsRunReader::releaseDevice()
{
  msp_timsData = nullptr;
  return true;
}

bool
TimsMsRunReader::acquireDevice()
{
  if(msp_timsData == nullptr)
    {
      initialize();
    }
  return true;
}


XicCoordSPtr
TimsMsRunReader::newXicCoordSPtrFromSpectrumIndex(std::size_t spectrum_index
                                                  [[maybe_unused]],
                                                  pappso::PrecisionPtr precision
                                                  [[maybe_unused]]) const
{
  throw ExceptionNotImplemented(QObject::tr("Not implemented %1 %2 %3")
                                  .arg(__FILE__)
                                  .arg(__FUNCTION__)
                                  .arg(__LINE__));
}

XicCoordSPtr
TimsMsRunReader::newXicCoordSPtrFromQualifiedMassSpectrum(
  const pappso::QualifiedMassSpectrum &mass_spectrum [[maybe_unused]],
  pappso::PrecisionPtr precision [[maybe_unused]]) const
{
  throw ExceptionNotImplemented(QObject::tr("Not implemented %1 %2 %3")
                                  .arg(__FILE__)
                                  .arg(__FUNCTION__)
                                  .arg(__LINE__));
}

TimsDataSp
TimsMsRunReader::getTimsDataSPtr()
{
  acquireDevice();
  return msp_timsData;
}


Trace
TimsMsRunReader::getTicChromatogram()
{
  // Use the Sqlite database to fetch the total ion current chromatogram (TIC
  // chromatogram).

  acquireDevice();

  // The time unit here is seconds, not minutes!!!
  return msp_timsData->getTicChromatogram();
}
