
/*******************************************************************************
 * Copyright (c) 2023 Olivier Langella <Olivier.Langella@u-psud.fr>.
 *
 * This file is part of the PAPPSOms++ library.
 *
 *     PAPPSOms++ is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     PAPPSOms++ is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with PAPPSOms++.  If not, see <http://www.gnu.org/licenses/>.
 *
 ******************************************************************************/

// cmake .. -DCMAKE_BUILD_TYPE=Debug  -DMAKE_TESTS=1  -DUSEPAPPSOTREE=1

//./tests/catch2-only-tests [MsRunReaderTims] -s


#include <catch2/catch_test_macros.hpp>
#include <catch2/catch_approx.hpp>
#include <catch2/matchers/catch_matchers_vector.hpp>


#include <iostream>

#include <pappsomspp/pappsoexception.h>
#include <pappsomspp/exception/exceptionnotpossible.h>
#include <pappsomspp/msfile/msfileaccessor.h>
#include <pappsomspp/msrun/output/mzxmloutput.h>
#include <pappsomspp/processing/uimonitor/uimonitorvoid.h>
#include <pappsomspp/msrun/private/timsframesmsrunreader.h>
#include <pappsomspp/processing/filters/filterpass.h>
#include <QDebug>
#include <QtCore>
#include <QFile>
#include <QtConcurrent>
#include "config.h"
// #include "common.h"

using namespace std;


TEST_CASE("Test MsRunReaderTims new API", "[MsRunReaderTims]")
{

  // Set the debugging message formatting pattern.
  qSetMessagePattern(QString("%{file}@%{line}, %{function}(): %{message}"));

#if USEPAPPSOTREE == 1

  SECTION("..::  test mz merge on tims ::..")

  {
    INFO("test on Bruker sample file");
    pappso::MsFileAccessor file_access_bruker(
      "/gorgone/pappso/versions_logiciels_pappso/bruker/"
      "200ngHeLaPASEF_2min_compressed.d",
      "");
    file_access_bruker.setPreferredFileReaderType(
      pappso::MsDataFormat::brukerTims, pappso::FileReaderType::tims);
    pappso::MsRunReaderSPtr msrunA01 =
      file_access_bruker.getMsRunReaderSPtrByRunId("", "runa01");

    REQUIRE(file_access_bruker.getFileReaderType() ==
            pappso::FileReaderType::tims);

    pappso::MsRunReaderQualifiedSpectrumList spectrum_list_reader;

    pappso::MsRunReadConfig config;
    config.setNeedPeakList(true);
    config.setMsLevels({1});
    config.setRetentionTimeStartInSeconds(2499);
    config.setRetentionTimeEndInSeconds(2499.5);

    qDebug();
    msrunA01.get()->readSpectrumCollection2(config, spectrum_list_reader);
    REQUIRE(spectrum_list_reader.getQualifiedMassSpectrumList().size() == 671);
    auto spectrum = spectrum_list_reader.getQualifiedMassSpectrumList().at(100);

    REQUIRE(spectrum.getMassSpectrumId().getSpectrumIndex() == 793893);
    REQUIRE(spectrum_list_reader.getQualifiedMassSpectrumList()
              .at(101)
              .getMassSpectrumId()
              .getSpectrumIndex() == 793893 + 1);

    auto spetrum_793893 = msrunA01.get()->qualifiedMassSpectrum(793893, true);
    REQUIRE(spetrum_793893.getMassSpectrumId().getSpectrumIndex() == 793893);

    REQUIRE(spectrum.getDtInMilliSeconds() ==
            Catch::Approx(spetrum_793893.getDtInMilliSeconds()));
    REQUIRE(
      spectrum
        .getParameterValue(pappso::QualifiedMassSpectrumParameter::OneOverK0)
        .toDouble() == Catch::Approx(1.4254876992));

    REQUIRE(
      spectrum_list_reader.getQualifiedMassSpectrumList()
        .at(101)
        .getParameterValue(pappso::QualifiedMassSpectrumParameter::OneOverK0)
        .toDouble() == Catch::Approx(1.4239005361));
    REQUIRE(
      spectrum
        .getParameterValue(pappso::QualifiedMassSpectrumParameter::OneOverK0)
        .toDouble() ==
      Catch::Approx(
        spetrum_793893
          .getParameterValue(pappso::QualifiedMassSpectrumParameter::OneOverK0)
          .toDouble()));


    REQUIRE(spetrum_793893.getMassSpectrumId().getNativeId().toStdString() ==
            "frame=1184 scan=100 index=793893");
    REQUIRE(spectrum.getMassSpectrumId().getNativeId().toStdString() ==
            "frame=1184 scan=100 index=793893");
    REQUIRE(spetrum_793893.getMassSpectrumCstSPtr().get()->xValues().size() ==
            373);
    REQUIRE_THAT(spetrum_793893.getMassSpectrumCstSPtr().get()->xValues(),
                 Catch::Matchers::Approx(
                   spectrum.getMassSpectrumCstSPtr().get()->xValues()));
    REQUIRE_THAT(spetrum_793893.getMassSpectrumCstSPtr().get()->yValues(),
                 Catch::Matchers::Approx(
                   spectrum.getMassSpectrumCstSPtr().get()->yValues()));
  }

  SECTION("..::  test mz merge on tims  restrict mobility range ::..")
  {
    INFO("test on Bruker sample file");
    pappso::MsFileAccessor file_access_bruker(
      "/gorgone/pappso/versions_logiciels_pappso/bruker/"
      "200ngHeLaPASEF_2min_compressed.d",
      "");
    file_access_bruker.setPreferredFileReaderType(
      pappso::MsDataFormat::brukerTims, pappso::FileReaderType::tims);
    pappso::MsRunReaderSPtr msrunA01 =
      file_access_bruker.getMsRunReaderSPtrByRunId("", "runa01");

    REQUIRE(file_access_bruker.getFileReaderType() ==
            pappso::FileReaderType::tims);

    pappso::MsRunReaderQualifiedSpectrumList spectrum_list_reader;

    pappso::MsRunReadConfig config;
    config.setNeedPeakList(true);
    config.setMsLevels({1});
    config.setRetentionTimeStartInSeconds(2499);
    config.setRetentionTimeEndInSeconds(2499.5);
    config.setParameterValue(pappso::MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mobility_index_begin,
                             200);
    config.setParameterValue(pappso::MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mobility_index_end,
                             300);
    config.setRetentionTimeEndInSeconds(2499.5);

    qDebug();
    msrunA01.get()->readSpectrumCollection2(config, spectrum_list_reader);
    REQUIRE(spectrum_list_reader.getQualifiedMassSpectrumList().size() == 101);
    auto spectrum = spectrum_list_reader.getQualifiedMassSpectrumList().at(2);

    REQUIRE(spectrum.getMassSpectrumId().getSpectrumIndex() == 793995);
    REQUIRE(spectrum_list_reader.getQualifiedMassSpectrumList()
              .at(3)
              .getMassSpectrumId()
              .getSpectrumIndex() == 793995 + 1);

    auto spetrum_793995 = msrunA01.get()->qualifiedMassSpectrum(793995, true);
    REQUIRE(spetrum_793995.getMassSpectrumId().getSpectrumIndex() == 793995);

    REQUIRE(spetrum_793995.getMassSpectrumId().getNativeId().toStdString() ==
            "frame=1184 scan=202 index=793995");
    REQUIRE(spectrum.getMassSpectrumId().getNativeId().toStdString() ==
            "frame=1184 scan=202 index=793995");
    REQUIRE(spetrum_793995.getMassSpectrumCstSPtr().get()->xValues().size() ==
            636);
    REQUIRE_THAT(spetrum_793995.getMassSpectrumCstSPtr().get()->xValues(),
                 Catch::Matchers::Approx(
                   spectrum.getMassSpectrumCstSPtr().get()->xValues()));
    REQUIRE_THAT(spetrum_793995.getMassSpectrumCstSPtr().get()->yValues(),
                 Catch::Matchers::Approx(
                   spectrum.getMassSpectrumCstSPtr().get()->yValues()));
  }

  SECTION("..::  test mz merge on tims  restrict mobility range + mz range::..")
  {
    INFO("test on Bruker sample file");
    pappso::MsFileAccessor file_access_bruker(
      "/gorgone/pappso/versions_logiciels_pappso/bruker/"
      "200ngHeLaPASEF_2min_compressed.d",
      "");
    file_access_bruker.setPreferredFileReaderType(
      pappso::MsDataFormat::brukerTims, pappso::FileReaderType::tims);
    pappso::MsRunReaderSPtr msrunA01 =
      file_access_bruker.getMsRunReaderSPtrByRunId("", "runa01");

    REQUIRE(file_access_bruker.getFileReaderType() ==
            pappso::FileReaderType::tims);

    pappso::MsRunReaderQualifiedSpectrumList spectrum_list_reader;

    pappso::MsRunReadConfig config;
    config.setNeedPeakList(true);
    config.setMsLevels({1});
    config.setRetentionTimeStartInSeconds(2499);
    config.setRetentionTimeEndInSeconds(2499.5);
    config.setParameterValue(pappso::MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mobility_index_begin,
                             200);
    config.setParameterValue(pappso::MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mobility_index_end,
                             300);
    config.setParameterValue(
      pappso::MsRunReadConfigParameter::TimsFramesMsRunReader_mz_begin, 300);
    config.setParameterValue(
      pappso::MsRunReadConfigParameter::TimsFramesMsRunReader_mz_end, 400);
    config.setRetentionTimeEndInSeconds(2499.5);

    qDebug();
    msrunA01.get()->readSpectrumCollection2(config, spectrum_list_reader);
    REQUIRE(spectrum_list_reader.getQualifiedMassSpectrumList().size() == 101);
    auto spectrum = spectrum_list_reader.getQualifiedMassSpectrumList().at(2);

    REQUIRE(spectrum.getMassSpectrumId().getSpectrumIndex() == 793995);
    REQUIRE(spectrum_list_reader.getQualifiedMassSpectrumList()
              .at(3)
              .getMassSpectrumId()
              .getSpectrumIndex() == 793995 + 1);

    auto spetrum_793995 = msrunA01.get()->qualifiedMassSpectrum(793995, true);
    REQUIRE(spetrum_793995.getMassSpectrumId().getSpectrumIndex() == 793995);

    REQUIRE(spetrum_793995.getMassSpectrumId().getNativeId().toStdString() ==
            "frame=1184 scan=202 index=793995");
    REQUIRE(spectrum.getMassSpectrumId().getNativeId().toStdString() ==
            "frame=1184 scan=202 index=793995");
    REQUIRE(spectrum.getMassSpectrumCstSPtr().get()->xValues().size() == 2);
    REQUIRE_THAT(spectrum.getMassSpectrumCstSPtr().get()->xValues(),
                 Catch::Matchers::Approx(
                   std::vector<double>({330.63218024, 331.5967379429})));
    REQUIRE_THAT(spectrum.getMassSpectrumCstSPtr().get()->yValues(),
                 Catch::Matchers::Approx(
                   std::vector<double>({30.1246063262, 56.1413117897})));
  }


  SECTION(
    "..::  test mz merge on tims  restrict mobility range + mz range + mz "
    "merge window::..")
  {
    INFO("test on Bruker sample file");
    pappso::MsFileAccessor file_access_bruker(
      "/gorgone/pappso/versions_logiciels_pappso/bruker/"
      "200ngHeLaPASEF_2min_compressed.d",
      "");
    file_access_bruker.setPreferredFileReaderType(
      pappso::MsDataFormat::brukerTims, pappso::FileReaderType::tims);
    pappso::MsRunReaderSPtr msrunA01 =
      file_access_bruker.getMsRunReaderSPtrByRunId("", "runa01");

    REQUIRE(file_access_bruker.getFileReaderType() ==
            pappso::FileReaderType::tims);

    pappso::MsRunReaderQualifiedSpectrumList spectrum_list_reader;

    pappso::MsRunReadConfig config;
    config.setNeedPeakList(true);
    config.setMsLevels({1});
    config.setRetentionTimeStartInSeconds(2499);
    config.setRetentionTimeEndInSeconds(2499.5);
    config.setParameterValue(pappso::MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mobility_index_begin,
                             200);
    config.setParameterValue(pappso::MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mobility_index_end,
                             300);
    config.setParameterValue(
      pappso::MsRunReadConfigParameter::TimsFramesMsRunReader_mz_begin, 300);
    config.setParameterValue(
      pappso::MsRunReadConfigParameter::TimsFramesMsRunReader_mz_end, 400);
    config.setParameterValue(pappso::MsRunReadConfigParameter::
                               TimsFramesMsRunReader_mz_index_merge_window,
                             1000);
    config.setRetentionTimeEndInSeconds(2499.5);

    qDebug();
    msrunA01.get()->readSpectrumCollection2(config, spectrum_list_reader);
    REQUIRE(spectrum_list_reader.getQualifiedMassSpectrumList().size() == 101);
    auto spectrum = spectrum_list_reader.getQualifiedMassSpectrumList().at(2);

    REQUIRE(spectrum.getMassSpectrumId().getSpectrumIndex() == 793995);
    REQUIRE(spectrum_list_reader.getQualifiedMassSpectrumList()
              .at(3)
              .getMassSpectrumId()
              .getSpectrumIndex() == 793995 + 1);

    auto spetrum_793995 = msrunA01.get()->qualifiedMassSpectrum(793995, true);
    REQUIRE(spetrum_793995.getMassSpectrumId().getSpectrumIndex() == 793995);

    REQUIRE(spetrum_793995.getMassSpectrumId().getNativeId().toStdString() ==
            "frame=1184 scan=202 index=793995");
    REQUIRE(spectrum.getMassSpectrumId().getNativeId().toStdString() ==
            "frame=1184 scan=202 index=793995");
    REQUIRE(spectrum.getMassSpectrumCstSPtr().get()->xValues().size() == 1);
    REQUIRE_THAT(
      spectrum.getMassSpectrumCstSPtr().get()->xValues(),
      Catch::Matchers::Approx(std::vector<double>({330.9439194641})));
    REQUIRE_THAT(spectrum.getMassSpectrumCstSPtr().get()->yValues(),
                 Catch::Matchers::Approx(std::vector<double>({86.2659181158})));
  }
#endif
}
